require "Automaten.rb"

##
# Author:: Bauer, Sebastian (mailto:sebastian.bauer@unibw.de)
# Author:: Jede, Alexander (mailto:alexander.jede@unibw.de)

##
# Diese Klasse reprsentiert die bergangstabelle eines LR(0)-Automaten.
# Deswegen arbeitet sie auch _NUR_ mit LR(0)-Automaten im DEA-Format.
# Jedem Zustand wird ein reprsentativer Name der Form K[i] mit i  |N zugeordnet.
# Wenn kein bergang zu einem anderem Zustand vorhanden ist, wird dies durch "-" dargestellt.
# 
class UTable
  ##
  # Der Konstruktor erwartet einen LR(0)-Automaten im DEA-Format.
  # Die bergangstabelle wird intern in einem zweidimensionalem Hash berechnet
  def initialize(dea)
    # Alle Zustnde in einem Hash speichern und ihnen einen "ordentlichen" Namen
    # nach folgendem Muster zuweissen: K[i] wobei i eine Nummer ist.
    # Eine Tablle ist ein Hash, mit Namen von Zustnden (das ist der Key), denen die bergnge
    # zugewiesen werden. Ein bergang ist ein Key (das Symbol) dem der neue Zustand (Name) als Wert
    # zugewiesen wird
    i = 1;
    @namehs = Hash.new
    @tabelle = Hash.new
    @zusths = Hash.new
    dea.zustaende_each{ |zust|
      @namehs["K#{i}"] = zust.to_s
      @zusths["K#{i}"] = zust
      @tabelle["K#{i}"] = Hash.new("-")
      i +=1
    }
    
    @uSymbole = Hash.new
    dea.uebergaenge_each{|konf|
      # Jede Konfiguration wird aufgespalten und die Information wird wie oben  beschrieben abgelegt
      @tabelle[@namehs.invert[konf[0].to_s]][konf[1]] = @namehs.invert[konf[2].to_s]
      @uSymbole[konf[1]] = konf[1]
    }
  end
  
  ##
  # Liefert die bergangstabelle als formatierten String.
  # Die Ausgangszustnde stehen Links in der Tabelle als Zeilen.
  # Die bergangssymbole stehen Oben als Spalten, die Zielzustnde
  # sind die Eintrge in dieser Matrix.
  def to_s
    symbole = @uSymbole.values
    maxZlaenge = (@namehs.size).to_s.length+1
    spaltenLaenge = Array.new
    # die Breite fr jede Spalte berechnen
    # richtet sich nach dem lngstem Eintrag in einer Spalte
    symbole.each{|symb|
      maxLaenge = 0
      @tabelle.each_value{|zeile|
        (zeile[symb].length > maxLaenge) ? (maxLaenge = zeile[symb].length) : () 
      }
      spaltenLaenge.push(maxLaenge)
    }
    
    # Kopfzeile
    erg = " " *(maxZlaenge+1)+"| "
    spaltenLaenge.each_index{|index|
      erg += (symbole[index].to_s) + " " *( spaltenLaenge[index] - symbole[index].to_s.length + 1)
    }
    erg += "\n"
    # Trennzeile
    erg += "-" *(erg.length-1) + "\n"
    #alle restlichen Zeilen ausgeben/hinzufgen
    @namehs.each_key{|zName|
      erg += zName + (" " *(maxZlaenge - zName.length + 1)) + "| "
      symbole.each_index{|index|
      folgeZust = (@tabelle[zName])[symbole[index]]
      erg += folgeZust + " " *(spaltenLaenge[index] - folgeZust.length + 1)
      }
      erg +="\n"
    }
    erg 
  end
  ##
  # Speichert die berganstabelle in eine Datei, dessen Pfad der Parameter ist.
  # Sollte die Datei nicht vorhanden sein, wird sie erstellt, sonst berschrieben
  def save(file)
	 require 'yaml'
	 File.open(file,"w"){|data|
	   YAML.dump(self,data)
	 }
  end
  
  ##
  # Es wird eine bergangstabelle aus dem bergebenem Pfad geladen und zurckgeliefert.
  def load(file)
	require 'yaml'
	load(File.open(file))
  end
  
end


##
# Diese Klasse reprsentiert die Aktionstabelle eines LR(0)-Automaten.
# Sie erbt von ATable und arbeitet deswegen auch _NUR_ mit LR(0)-Automaten im DEA-Format.
# Der Aktionstabelle kann entnommen werden, abhngig vom aktuellen Zustand, 
# welche Aktion durch gefhrt werden kann.
# Denkbar wren:
# * shift : zum Lesen eines Zeichen
# * reduce : Reduktion einer Regel
# * error : wenn kein bergang mglich ist.
# Dazu mehr im Buch: "Syntaxbassierte Programmierwerkzeuge" von Lothar Schmitz, Seite 178.

class ATable < UTable

  ##
  # Der Konstruktor erwartet einen LR(0)-Automaten im DEA-Format.
  # Die Aktionstabelle wird intern in einem zweidimensionalem Hash berechnet.
  # 
  def initialize(dea)
    # Alle Zustnde in einem Hash speichern und ihnen einen "ordentlichen" Namen
    # nach folgendem Muster zuweissen: K[i] wobei i eine Nummer ist.
    # Eine Tablle ist ein Hash, mit Namen von Zustnden (das ist der Key), denen die Aktionen
    # zugewiesen werden. Eine Aktion kann sein (s.S. 155 u. 178):
    # * Reduktion
    # * Shift
    # * Error
    super
    @conflictArray = Array.new
    # in der Tabelle alle Eintrge durchlaufen und alle NonTerminale Spalten entfernen
    @tabelle.each{|zKey,zeile|
      hasshift = false
     hasreduk = false
      @uSymbole.each_key{|symb|
        countReduk=0
        if !symb.terminal
          # Wir haben ein NonTerminal
          # also aus der Tabelle entfernen und zusttzlich aus der Tabelle uSymb entfernen
          zeile.delete(symb)
          @uSymbole.delete(symb)
        else
          # Wenn es ein begang zu einem Zustand ist, dann wird es ein Shift
          if zeile[symb] != "-"
            zeile[symb] = "shift"
            hasshift = true
          else 
            # Berechnen, ob es ein Reduce oder ein Error ist
            # iterieren durch die Items eines Zustands
            @zusths[zKey].each{|item|
              # Is das vorletze Zeichen ein Punkt(:=46), dann ist es eine Reduktion
              if item.to_s[-2] == 46
                countReduk += 1
                hasreduk = true
                zeile[symb] = "red#{item.to_s}"
              end
            }
          end
        end
      if hasshift == true and hasreduk 
        @conflictArray.push "#{zKey}: shift-reduce-conflict"
      elsif countReduk > 1 
        @conflictArray.push "#{zKey}: reduce-reduce-conflict"
      end
      }
      zeile.default= "error"
      
    }    
  end
  
  ##
  # Liefert den Array mit den Konfliktmeldungen. Diese liegen in einem String vor.
  def getConflicts
   @conflictArray
  end
  
  ##
  # Liefert einen boolschen Wert, ob der LR(0)-Automat konfliktfrei ist.
  def isConflictfree
   @conflictArray.empty?
  end
  
  ##
  # Liefert einen String, in dem alle Konflikte aufgelistet sind.
  # Sollte kein Konflikt vorliegen, steht dieses im String drin.
  def printConflicts
    if isConflictfree 
      return "Der Automat ist konfliktfrei."
    else
    confStr = String.new
      @conflictArray.each { |conflict|
        confStr += conflict
      } 
     end
  end
end

=begin
  a = LR_0.new(Grammar.new("S -> b S d\n | b S c S d \n | a"))
  puts (UTable.new a).to_s
  puts "\n" + (tab = ATable.new a).to_s
  puts tab.printConflicts
=end
b = LR_0.new(g = Grammar.new("S -> T + S\n | T\n\nT -> F * T\n | F \n \nF -> id \n|( S )"))
puts (UTable.new b).to_s
puts "\n" + (tab = ATable.new b).to_s
puts tab.printConflicts
=begin
  inp = gets.chop

  while inp != ""
    puts "#{inp} = #{eval inp}"
    inp = gets.chop
  end
=end
    
#end